
#include <map>
#include "precompiled_libcef.h"
#include "base/string_util.h"
#include "base/worker_pool.h"
#include "base/lazy_instance.h"
#include "base/message_loop.h"
#include "net/url_request/url_request_job.h"
#include "net/url_request/url_request.h"
#include "net/url_request/url_request_filter.h"
#include "net/base/completion_callback.h"
#include "net/base/io_buffer.h"
#include "net/base/upload_data.h"
#include "net/http/http_util.h"
#include "googleurl/src/url_util.h"

#include "cef_scheme.h"
#include "tracker.h"
#include "context.h"
#include "request_impl.h"

// Memory manager.

base::LazyInstance<CefTrackManager> g_scheme_tracker(base::LINKER_INITIALIZED);

class TrackBase : public CefTrackObject
{
public:
	TrackBase(CefBase* base) { base_ = base; }

protected:
	CefRefPtr<CefBase> base_;
};

static void TrackAdd(CefTrackObject* object)
{
	g_scheme_tracker.Pointer()->Add(object);
}

static void TrackDelete(CefTrackObject* object)
{
	g_scheme_tracker.Pointer()->Delete(object);
}

// a URLRequestFilter clone
// feature removed: AddUrlHandler RemoveUrlHandler
// feature changed: AddHostnameHandler
// The new AddHostnameHandler function allows an extra handler factory
// pointer to be stored. FindRequestHandlerFactory can be used to obtain
// this factory pointer, which is used by CefUrlRequestJob's factory
// method
class CefUrlRequestFilter {
public:
	// scheme,hostname -> ProtocolFactory
	typedef std::map<std::pair<std::string, std::string>,
		URLRequest::ProtocolFactory*> HostnameHandlerMap;
	typedef std::map<std::pair<std::string, std::string>,
		CefSchemeHandlerFactory*> HostnameHandlerFactoryMap;

	// Singleton instance for use.
	static CefUrlRequestFilter* GetInstance()
	{
		if (!shared_instance_)
			shared_instance_ = new CefUrlRequestFilter;
		return shared_instance_;
	}

	static URLRequestJob* Factory(URLRequest* request,
		const std::string& scheme)
	{
		// Returning null here just means that the built-in handler will be used.
		return GetInstance()->FindRequestHandler(request, scheme);
	}

	void AddHostnameHandler(const std::string& scheme,
		const std::string& hostname,
		CefSchemeHandlerFactory* handlerfactory,
		URLRequest::ProtocolFactory* factory)
	{
		hostname_handler_map_[make_pair(scheme, hostname)] = factory;
		hostname_handler_factory_map_[make_pair(scheme, hostname)] = handlerfactory;

		// Register with the ProtocolFactory.
		URLRequest::RegisterProtocolFactory(scheme,
			&CefUrlRequestFilter::Factory);
	}
	void RemoveHostnameHandler(const std::string& scheme,
		const std::string& hostname)
	{
		HostnameHandlerMap::iterator iter =
			hostname_handler_map_.find(make_pair(scheme, hostname));
		DCHECK(iter != hostname_handler_map_.end());
		HostnameHandlerFactoryMap::iterator iterf =
			hostname_handler_factory_map_.find(make_pair(scheme, hostname));
		DCHECK(iterf != hostname_handler_factory_map_.end());

		hostname_handler_map_.erase(iter);
		hostname_handler_factory_map_.erase(iterf);
	}

	// Clear all the existing URL handlers and unregister with the
	// ProtocolFactory.  Resets the hit count.
	void ClearHandlers()
	{
		// Unregister with the ProtocolFactory.
		std::set<std::string> schemes;
		for (HostnameHandlerMap::const_iterator i = hostname_handler_map_.begin();
			i != hostname_handler_map_.end(); ++i) {
				schemes.insert(i->first.first);
		}
		for (std::set<std::string>::const_iterator scheme = schemes.begin();
			scheme != schemes.end(); ++scheme) {
				URLRequest::RegisterProtocolFactory(*scheme, NULL);
		}

		hostname_handler_map_.clear();
		hit_count_ = 0;
	}

	CefSchemeHandlerFactory* FindRequestHandlerFactory(URLRequest* request,
		const std::string& scheme)
	{
		CefSchemeHandlerFactory* factory = NULL;
		if (request->url().is_valid()) {
			// Check the hostname map first.
			const std::string& hostname = request->url().host();

			HostnameHandlerFactoryMap::iterator i =
				hostname_handler_factory_map_.find(make_pair(scheme, hostname));
			if (i != hostname_handler_factory_map_.end())
				factory = i->second;

		}
		return factory;
	}

	// Returns the number of times a handler was used to service a request.
	int hit_count() const { return hit_count_; }

protected:
	CefUrlRequestFilter() : hit_count_(0) { }

	// Helper method that looks up the request in the url_handler_map_.
	URLRequestJob* FindRequestHandler(URLRequest* request,
		const std::string& scheme)
	{
		URLRequestJob* job = NULL;
		if (request->url().is_valid()) {
			// Check the hostname map first.
			const std::string& hostname = request->url().host();

			HostnameHandlerMap::iterator i =
				hostname_handler_map_.find(make_pair(scheme, hostname));
			if (i != hostname_handler_map_.end())
				job = i->second(request, scheme);

		}
		if (job) {
			DLOG(INFO) << "URLRequestFilter hit for " << request->url().spec();
			hit_count_++;
		}
		return job;
	}

	// Maps hostnames to factories.  Hostnames take priority over URLs.
	HostnameHandlerMap hostname_handler_map_;
	HostnameHandlerFactoryMap hostname_handler_factory_map_;

	int hit_count_;

private:
	// Singleton instance.
	static CefUrlRequestFilter* shared_instance_;

	DISALLOW_EVIL_CONSTRUCTORS(CefUrlRequestFilter);
};

CefUrlRequestFilter* CefUrlRequestFilter::shared_instance_ = NULL;


class CefUrlRequestJob : public URLRequestJob {
public:
	CefUrlRequestJob(URLRequest* request, CefSchemeHandler* handler)
		: URLRequestJob(request),
		url_(request->url()),
		handler_(handler),
		remaining_bytes_(0) {	}

	virtual ~CefUrlRequestJob(){}

	virtual void Start()
	{
		handler_->Cancel();
		// Continue asynchronously.
		DCHECK(!async_resolver_);
		async_resolver_ = new AsyncResolver(this);
		WorkerPool::PostTask(FROM_HERE, NewRunnableMethod(
			async_resolver_.get(), &AsyncResolver::Resolve, url_), true);
		return;
	}
	virtual void Kill()
	{
		if (async_resolver_) {
			async_resolver_->Cancel();
			async_resolver_ = NULL;
		}

		URLRequestJob::Kill();
	}
	virtual bool ReadRawData(net::IOBuffer* dest, int dest_size, int *bytes_read)
	{
		DCHECK_NE(dest_size, 0);
		DCHECK(bytes_read);

		// When remaining_bytes_>=0, it means the handler knows the content size
		// before hand. We continue to read until 
		if (remaining_bytes_>=0)
		{
			if (remaining_bytes_ < dest_size)
				dest_size = static_cast<int>(remaining_bytes_);

			// If we should copy zero bytes because |remaining_bytes_| is zero, short
			// circuit here.
			if (!dest_size) {
				*bytes_read = 0;
				return true;
			}

			// remaining_bytes > 0
			bool rv = handler_->ReadResponse(dest->data(), dest_size, bytes_read);
			remaining_bytes_ -= *bytes_read;
			if(!rv)
				// handler indicated no further data to read
				*bytes_read = 0;
			return true;
		}
		// The handler returns -1 for GetResponseLength, this means the handler
		// doesn't know the content size before hand. We do basically the same
		// thing, except for checking the return value for handler_->ReadResponse,
		// which is an indicator for no further data to be read.
		else
		{
			bool rv = handler_->ReadResponse(dest->data(), dest_size, bytes_read);
			if(!rv)
				// handler indicated no further data to read
				*bytes_read = 0;
			return true;
		}
	}
	virtual bool IsRedirectResponse(GURL* location, int* http_status_code){ return false; }
	virtual bool GetContentEncodings(
		std::vector<Filter::FilterType>* encoding_types)
	{
		DCHECK(encoding_types->empty());

		return !encoding_types->empty();
	}
	virtual bool GetMimeType(std::string* mime_type) const
	{
		DCHECK(request_);
		// call handler to get mime type
		*mime_type = WideToUTF8(handler_->GetMimeType()).c_str();
		return true;
	}
	virtual void SetExtraRequestHeaders(const std::string& headers)
	{
	}

	static URLRequestJob* Factory(URLRequest* request, const std::string& scheme)
	{
		if(CefSchemeHandlerFactory* factory = 
			CefUrlRequestFilter::GetInstance()->FindRequestHandlerFactory(request, scheme))
		{
			OutputDebugStringA("Calling URLRequestJob Factory method\n");
			CefRefPtr<CefSchemeHandler> handler = factory->Create();
			return new CefUrlRequestJob(request, handler.get());
		}
		else
			// should never be reached.
			return NULL;
	}

	CefRefPtr<CefSchemeHandler> handler_;

protected:
	GURL url_;

private:
	void DidResolve(const GURL& url)
	{
		async_resolver_ = NULL;

		// We may have been orphaned...
		if (!request_)
			return;

		remaining_bytes_ = handler_->GetResponseLength();
		if(remaining_bytes_>0)
		set_expected_content_size(remaining_bytes_);
		NotifyHeadersComplete();
	}

	int64 remaining_bytes_;
	std::string m_response;

	class AsyncResolver :
		public base::RefCountedThreadSafe<AsyncResolver> {
	public:
		explicit AsyncResolver(CefUrlRequestJob* owner)
			: owner_(owner), owner_loop_(MessageLoop::current()) {
		}

		void Resolve(const GURL& url) {
			AutoLock locked(lock_);
			//////////////////////////////////////////////////////////////////////////
			// safe to perform long operation here
			CefRefPtr<CefRequest> req(CefRequest::CreateRequest());
			req->SetURL(UTF8ToWide(url.spec()));
			req->SetMethod(
				UTF8ToWide(owner_->request()->method()));

			// check to see if we have post data
			if(net::UploadData *data = owner_->request()->get_upload())
			{
				CefRefPtr<CefPostData> postdata(CefPostData::CreatePostData());
				static_cast<CefPostDataImpl*>(postdata.get())->Set(*data);
				req->SetPostData(postdata);
			}
			// handler should complete content generation in ProcessRequest
			owner_->handler_->Cancel();
			owner_->handler_->ProcessRequest(req);
			//////////////////////////////////////////////////////////////////////////
			if (owner_loop_) {
				owner_loop_->PostTask(FROM_HERE, NewRunnableMethod(
					this, &AsyncResolver::ReturnResults, url));
			}
		}

		void Cancel() {
			owner_->handler_->Cancel();

			owner_ = NULL;

			AutoLock locked(lock_);
			owner_loop_ = NULL;
		}

	private:
		void ReturnResults(const GURL& url) {
			if (owner_)
				owner_->DidResolve(url);
		}

		CefUrlRequestJob* owner_;

		Lock lock_;
		MessageLoop* owner_loop_;
	};

	friend class AsyncResolver;
	scoped_refptr<AsyncResolver> async_resolver_;

	DISALLOW_COPY_AND_ASSIGN(CefUrlRequestJob);
};


class SchemeRequestJobWrapper {
public:
	SchemeRequestJobWrapper(const std::string& scheme_name,
		const std::string& host_name,
		CefSchemeHandlerFactory* factory)
	: factory_(factory), scheme_name_(scheme_name), host_name_(host_name)
	{
		// The reference will be released when the application exits.
		TrackAdd(new TrackBase(factory));
	}

	void RegisterScheme()
	{
		// we need to store the pointer of this handler because
		// we can't pass it as a parameter to the factory method
		CefUrlRequestFilter* filter = CefUrlRequestFilter::GetInstance();
		filter->AddHostnameHandler(scheme_name_, host_name_, factory_, 
			&CefUrlRequestJob::Factory);
	}

	void AddRef() {}
	void Release() {}

private:
	CefSchemeHandlerFactory* factory_;
	std::string scheme_name_;
	std::string host_name_;
};

bool CefRegisterScheme(	const std::wstring& scheme_name,
											 const std::wstring& host_name,
											 CefRefPtr<CefSchemeHandlerFactory> factory)
{
	// Verify that the context is already initialized
	if(!CefUrlRequestFilter::GetInstance())
		return false;

	SchemeRequestJobWrapper* wrapper = new SchemeRequestJobWrapper(WideToUTF8(scheme_name),
		WideToUTF8(host_name), factory);

	PostTask(FROM_HERE, NewRunnableMethod(wrapper,
		&SchemeRequestJobWrapper::RegisterScheme));
	return true;
}
